
import asyncio


try:
    from qtpy import QtWidgets
    has_qt = True
    qt_error = False
        
except Exception as e:
    has_qt = False
    qt_error = e
    QtWidgets = None


try:
    from qtpy.QtWebEngineWidgets import QWebEngineView
    has_webview = True
        
except Exception as e:
    has_webview = False
    web_view_error = e


class GraphicalBackend:


    def __init__(self, parent=None) -> None:
        self.parent = parent
        self.spawned_app = None


    def __enter__(self):
        global has_qt
        assert has_qt == True, f"You cannot run with a Qt Backend if no QT Backend is installed. Please install PyQT5 {str(has_qt)}"

        if QtWidgets.QApplication.instance() is None:
            # if it does not exist then a QApplication is created
            self.spawned_app = QtWidgets.QApplication([])

        return self


    def __exit__(self, *args, **kwargs):
        if self.spawned_app: self.spawned_app.exit()


    async def __aenter__(self):
        return self.__enter__()

    async def __aexit__(self, type, value, traceback):
        
        if type is not None:
            if issubclass(type, asyncio.CancelledError): 
                print("Raising cancellation")
                raise type(value).with_traceback(traceback)

            if issubclass(type, Exception):
                print(f"Raising exceütopm {type} {value} {traceback}")
                raise type(value).with_traceback(traceback)


            raise type(value).with_traceback(traceback)


from abc import abstractmethod
from herre.loop import loopify
from herre.wards.variables import parse_variables
from herre.auth import HerreClient
import asyncio

class WardException(Exception):
    pass

class WardMeta(type):
    """
    
    """



    def __init__(self, name, bases, attrs):
        super(WardMeta, self).__init__(name, bases, attrs)
        if attrs["__qualname__"] != "BaseWard":
            # This gets allso called for our Baseclass which is abstract
            meta = attrs["Meta"] if "Meta" in attrs else None
            assert meta is not None, f"Please provide a Meta class in your Arnheim Model {name}"

            try:
                if meta.abstract:
                    return
            except:
                pass

            register = getattr(meta, "register", True)

            if register:
                from herre.wards.registry import get_ward_registry
                key = getattr(meta, "key", None)
                assert key is not None, f"Please provide key in your Meta class to register the Ward {attrs['__qualname__']}, or specifiy register=False"
                get_ward_registry().register_ward(key, self)



class BaseQuery(object):
    """Query
    
    A query is an instructional set of how to retrieve data from a Server through a ward. A Ward takes a query populates it with credentials through
    herre and send its to its respective instance. 

    We ship with a Standard GraphQL Query.

    This can be used as a mixin for typesaftey
    
    
    
    """


class BaseWard(metaclass=WardMeta):
    """Ward

    Wards are connectors between Models and there Corresponding Endpoints (Servers). They are automatically registered in a common ward registry
    so models can just specify the ward they want to use by referencing it in their Meta 

    Args:
        BaseModel ([type]): [description]
        metaclass ([type], optional): [description]. Defaults to ModelMeta.

    Raises:
        NotImplementedError: [description]
        NotImplementedError: [description]

    Returns:
        [type]: [description]
    """
    id: str

    def __init__(self, herre: HerreClient) -> None:
        self.herre = herre
        self.connected = False
        self.transcript = None
        super().__init__()

    @abstractmethod
    async def run(self, query: BaseQuery, variables: dict):
        raise NotImplementedError("Your Ward must overwrite run")

    @abstractmethod
    async def connect(self):
        raise NotImplementedError("Your Ward must overwrite run")

    @abstractmethod
    async def disconnect(self):
        raise NotImplementedError("Your Ward must overwrite run")


    async def negotiate(self):
        """Negotiation is a step before launching the first query to your backend service,
        it allows for initial configurations to be transfer
        """
        pass


    async def _run_async(self, query: BaseQuery, variables: dict):
        assert isinstance(query, BaseQuery), "Query must be of type BaseQuery"
        if not self.connected:
            await self._connect()
        return await self.run(query, await parse_variables(variables))


    def _run_sync(self, query: BaseQuery, variables: dict):
        return loopify(self._run_async(query, variables))



    async def _disconnect(self):        
        await self.disconnect()
        self.transcript = None
        self.connected = False

    async def _connect(self):        
        await self.connect()
        self.transcript = await self.negotiate()
        self.connected = True
        


    class Meta:
        abstract = True








from typing import Dict, List, Optional
from pydantic import BaseModel
from enum import Enum

from typing import ForwardRef



class TypeKind(str, Enum):
    SCALAR = "SCALAR"
    OBJECT = "OBJECT"
    LIST = "LIST"
    INTERFACE = "INTERFACE"
    UNION = "UNION"
    ENUM = "ENUM"
    INPUT_OBJECT = "INPUT_OBJECT"
    NON_NULL = "NON_NULL"


Type = ForwardRef("Type")
InputValue = ForwardRef("InputValue")
Field = ForwardRef("Field")

class InputValue(BaseModel):
    name: Optional[str]
    description: Optional[str]
    type: Optional[Type]
    defaultValue: Optional[str]


class Field(BaseModel):
    name: Optional[str]
    description:  Optional[str]
    args: Optional[List[InputValue]]
    type: Optional[Type]
    isDeprecated: Optional[bool]
    deprecationReason: Optional[bool]


class Type(BaseModel):
    kind: Optional[TypeKind]
    description: Optional[str]
    name: Optional[str]
    fields: Optional[List[Field]]
    interfaces:  Optional[List[Type]]
    inputFields:  Optional[List[InputValue]]
    ofType: Optional[Type]



Type.update_forward_refs()
InputValue.update_forward_refs()
Field.update_forward_refs()


class Schema(BaseModel):
    types: Optional[List[Type]]
    queryType: Optional[Type]



from pydantic.main import BaseModel
import yaml
import os


class ConfigError(Exception):
    pass

class BaseConfig(BaseModel):
    """Container for the Configuration options
    parsed either from Env or from the yaml file

    Abstract:
        __group (str): Needs to implement for the Parsing

    Args:
        BaseModel ([type]): [description]

    Returns:
        [type]: [description]
    """
    _group: str = None


    @classmethod
    def from_file(cls, file_path, **overrides):
        assert cls._group is not None, "Please specifiy your parent group in your Config Class to access it from a file "
        assert os.path.isfile(file_path), f"File {file_path} does not exist"
        
        with open(file_path,"r") as file:
            config = yaml.load(file, Loader=yaml.FullLoader)

            try:
                for subgroup in cls._group.split("."):
                    config = config[subgroup]

                config.update(**overrides)
                return cls(**config)
            except KeyError as e:
                raise ConfigError(f"Couldn't load {cls._group} Group from {file_path}: Subgroup {subgroup} does not exist in {config}") from e


from typing import Union
import contextvars

try:
    from rich.console import Console as RichConsole
    console_instance = RichConsole()

except ImportError as e:
    console_instance = None


console = contextvars.ContextVar("console", default=console_instance)

def get_current_console():
    return console.get() 


class Console():

    def __init__(self) -> None:
        self.console = RichConsole()


    def __enter__(self):
        console.set(self)
        return self


    def __exit__(self,*args, **kwargs):
        pass



import logging

try:
    from rich.logging import RichHandler
    StreamHandler = RichHandler
except ImportError as e:
    from logging import StreamHandler


def setLogging(log_level, log_stream=False, log_file=True):
    

    
    logger = logging.getLogger()
    logger.setLevel(log_level)

    if log_stream:
        stream = StreamHandler()
        logging.root.setLevel(log_level)
        stream.setLevel(log_level)

        logger.addHandler(stream)

    if log_file:
        file = logging.FileHandler("logs.txt")
        logging.root.setLevel(log_level)
        file.setLevel(log_level)

        logger.addHandler(file)



setLogging("INFO", log_stream=True)







from herre.jupy import in_notebook
from herre.grants.code.app import AuthorizationCodeGrant
from herre.grants.backend.app import BackendGrant
from herre.grants.backend.app import BackendGrant
from herre.config.model import GrantType, HerreConfig
import asyncio 
from threading import Thread
import os
import logging

logger = logging.getLogger(__name__)

class HerreError(Exception):
    pass

def newloop(loop):
    asyncio.set_event_loop(loop)
    loop.run_forever()




class HerreClient:

    def __init__(self,
        config_path = "bergen.yaml",
        username: str = None,
        password: str = None,
        allow_insecure = False,
        force_async = False,
        force_sync = False,
        register = True,
        **overrides,
    ) -> None:
        """ Creates A Herre Client

        Args:
            config_path (str, optional): [description]. Defaults to "bergen.yaml".
            username (str, optional): [description]. Defaults to None.
            password (str, optional): [description]. Defaults to None.
            allow_insecure (bool, optional): [description]. Defaults to False.
            in_sync (bool, optional): Should we force an in_sync modus if an event loop is already running. Loop will be send to another thread. Defaults to True.

        Raises:
            HerreError: [description]
        """

        loop = asyncio.get_event_loop()
        if not force_async and loop.is_running() and in_notebook() :
            # Jupyter handles the eventloop a bit erradically put has top_level await, so we
            # can either have sync or non_sync code, as long as await is not a standard
            # we will use sync_mode for jupyter

            # If we are already in an event loop we are creating a new eventloop somewhere else
            thread_loop = asyncio.new_event_loop()
            t = Thread(target=newloop, args=(thread_loop,))
            t.start()
            logger.info("Running in Seperate Thread")
            self.loop = thread_loop
            self.sync_mode = True
        else:
            self.loop = loop
            self.sync_mode = False


        self.config = HerreConfig.from_file(config_path, **overrides)

        if self.config.authorization_grant_type == GrantType.CLIENT_CREDENTIALS:
            self.grant = BackendGrant(self.config)

        elif self.config.authorization_grant_type == GrantType.AUHORIZATION_CODE:
            self.grant = AuthorizationCodeGrant(self.config)

        else:
            raise HerreError("Unknown GrantType")

        if register:
            set_current_herre(self)


    
    async def login(self, username: str = None, password: str = None):
        await self.grant.login(username=username, password=password)

    
    async def refresh(self):
        assert self.grant.can_refresh, "Grant is not able to refresh. Please login again"
        await self.grant.refresh()


    @property
    def logged_in(self):
        return self.grant.logged_in

    @property
    def headers(self, **additionals):
        assert self.grant.access_token is not None, "Access token is not set yet, please login?"
        return {"Authorization": f"Bearer {self.grant.access_token}", **additionals}




CURRENT_HERRE = None

def get_current_herre():
    global CURRENT_HERRE
    if not CURRENT_HERRE:
        CURRENT_HERRE = HerreClient()
    return CURRENT_HERRE

def set_current_herre(herre):
    global CURRENT_HERRE
    CURRENT_HERRE = herre


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['herre',
 'herre.config',
 'herre.console',
 'herre.grants',
 'herre.grants.backend',
 'herre.grants.code',
 'herre.grants.code.widgets',
 'herre.wards']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'QtPy>=1.11.1,<2.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'requests-oauthlib>=1.3.0,<2.0.0',
 'rich>=10.9.0,<11.0.0']

extras_require = \
{'jupyter': ['jupyter>=1.0.0,<2.0.0'],
 'pyqt': ['PyQt5>=5.15.4,<6.0.0', 'PyQtWebEngine>=5.15.4,<6.0.0']}

setup_kwargs = {
    'name': 'herre',
    'version': '0.1.0',
    'description': 'The governing package for modules of the arnheim framework, it provides core features like wards, the event loop and authentication with an oauth provider',
    'long_description': '# Herre\n\n### Idea\n\nHerre is the core library to provide the groundwork for every module of the arkitekt framework,\n\n\n \n### Prerequisites\n\nHerre only works with a running Oauth Instance (in your network or locally for debugging).\n\n### Usage\n\nIn order to initialize the Client you need to connect it as a Valid Application with your Arnheim Instance\n\n```python\nclient = Bergen(host="p-tnagerl-lab1",\n    port=8000,\n  client_id="APPLICATION_ID_FROM_ARNHEIM", \n  client_secret="APPLICATION_SECRET_FROM_ARNHEIM",\n  name="karl",\n)\n```\n\nIn your following code you can simple query your data according to the Schema of the Datapoint\n\n```python\nfrom bergen.schema import Node\n\nnode = Node.objects.get(id=1)\nprint(node.name)\n\n```\n\n## Access Data from different Datapoints\n\nThe Arnheim Framework is able to provide data from different Data Endpoints through a commong GraphQL Interface\n. This allows you to access data from various different storage formats like Elements and Omero and interact without\nknowledge of their underlying api.\n\nEach Datapoint provides a typesafe schema. Arnheim Elements provides you with an implemtation of that schema.\n\n## Provide a Template for a Node\n\nDocumentation neccesary\n\n\n### Testing and Documentation\n\nSo far Bergen does only provide limitedunit-tests and is in desperate need of documentation,\nplease beware that you are using an Alpha-Version\n\n\n### Build with\n\n- [Arnheim](https://github.com/jhnnsrs/arnheim)\n- [Pydantic](https://github.com/jhnnsrs/arnheim)\n\n',
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
